@@ -32,10 +32,12 @@
 #include <stdio.h>
 #include <stdarg.h>
 #include <string.h>
+#include<sys/stat.h>
 #include <atk/atk.h>
 
 #include <droute/droute.h>
 #include <gmodule.h>
+#include <glib/gi18n.h>
 
 #include "bridge.h"
 #include "event.h"
@@ -149,6 +151,7 @@ spi_atk_bridge_get_bus (void)
   else
     {
       bus = dbus_connection_open (data, &error);
+      XFree (data);
       if (!bus)
         {
           g_warning ("AT-SPI: Couldn't connect to bus: %s\n", error.message);
@@ -173,25 +176,9 @@ set_reply (DBusPendingCall *pending, void *user_data)
     void **replyptr = (void **)user_data;
 
     *replyptr = dbus_pending_call_steal_reply (pending);
+  dbus_pending_call_unref (pending);
 }
 
-static DBusMessage *
-send_and_allow_reentry (DBusConnection *bus, DBusMessage *message, DBusError *error)
-{
-    DBusPendingCall *pending;
-    DBusMessage *reply = NULL;
-
-    if (!dbus_connection_send_with_reply (bus, message, &pending, -1))
-    {
-        return NULL;
-    }
-    dbus_pending_call_set_notify (pending, set_reply, (void *)&reply, NULL);
-    while (!reply)
-    {
-      if (!dbus_connection_read_write_dispatch (bus, -1)) return NULL;
-    }
-    return reply;
-}
 /*---------------------------------------------------------------------------*/
 
 static void
@@ -243,10 +230,7 @@ get_registered_event_listeners (SpiBridge *app)
     }
   dbus_message_iter_init (reply, &iter);
   dbus_message_iter_recurse (&iter, &iter_array);
-  /* TODO: This is bad. Need to determine that the array is non-empty,
-     so that we don't initially read a value rom it in that case, but using
-     a deprecated function. */
-  if (dbus_message_iter_get_array_len (&iter_array) > 0) do
+  while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
     {
       char *bus_name, *event;
       dbus_message_iter_recurse (&iter_array, &iter_struct);
@@ -254,8 +238,8 @@ get_registered_event_listeners (SpiBridge *app)
       dbus_message_iter_next (&iter_struct);
       dbus_message_iter_get_basic (&iter_struct, &event);
       add_event (bus_name, event);
+      dbus_message_iter_next (&iter_array);
     }
-  while (dbus_message_iter_next (&iter_array));
   dbus_message_unref (reply);
 }
 
@@ -266,7 +250,8 @@ register_reply (DBusPendingCall *pending, void *user_data)
   SpiBridge *app = user_data;
   DBusMessage *message;
 
-    reply = dbus_pending_call_steal_reply (pending);
+  reply = dbus_pending_call_steal_reply (pending);
+  dbus_pending_call_unref (pending);
   if (reply)
     {
       gchar *app_name, *obj_path;
@@ -274,7 +259,6 @@ register_reply (DBusPendingCall *pending, void *user_data)
       if (strcmp (dbus_message_get_signature (reply), "(so)") != 0)
         {
           g_warning ("AT-SPI: Could not obtain desktop path or name\n");
-printf("sig: %s\n", dbus_message_get_signature(reply));
         }
       else
         {
@@ -306,6 +290,7 @@ register_application (SpiBridge * app)
   DBusMessageIter iter;
   DBusError error;
   DBusPendingCall *pending;
+  const int max_addr_length = 128; /* should be long enough */
 
   dbus_error_init (&error);
 
@@ -322,7 +307,8 @@ register_application (SpiBridge * app)
   dbus_message_iter_init_append (message, &iter);
   spi_object_append_reference (&iter, app->root);
   
-    if (!dbus_connection_send_with_reply (app->bus, message, &pending, -1))
+    if (!dbus_connection_send_with_reply (app->bus, message, &pending, -1)
+        || !pending)
     {
         return FALSE;
     }
@@ -332,6 +318,18 @@ register_application (SpiBridge * app)
   if (message)
     dbus_message_unref (message);
 
+  /* could this be better, we accept some amount of race in getting the temp name*/
+  /* make sure the directory exists */
+  mkdir ("/tmp/at-spi2/", S_IRWXU|S_IRWXG|S_IRWXO|S_ISVTX);
+  chmod ("/tmp/at-spi2/", S_IRWXU|S_IRWXG|S_IRWXO|S_ISVTX);
+  app->app_bus_addr = g_malloc(max_addr_length * sizeof(char));
+#ifndef DISABLE_P2P
+  sprintf (app->app_bus_addr, "unix:path=/tmp/at-spi2/socket-%d-%d", getpid(),
+           rand());
+#else
+  app->app_bus_addr [0] = '\0';
+#endif
+
   return TRUE;
 }
 
@@ -508,6 +506,50 @@ install_plug_hooks ()
   socket_class->embed = socket_embed_hook;
 }
 
+static void
+new_connection_cb (DBusServer *server, DBusConnection *con, void *data)
+{
+  GList *new_list;
+
+  dbus_connection_ref(con);
+  dbus_connection_setup_with_g_main(con, NULL);
+  droute_intercept_dbus (con);
+  droute_context_register (spi_global_app_data->droute, con);
+
+  new_list = g_list_append (spi_global_app_data->direct_connections, con);
+  if (new_list)
+    spi_global_app_data->direct_connections = new_list;
+}
+
+static int
+setup_bus (void)
+{
+#ifndef DISABLE_P2P
+  DBusServer *server;
+  DBusError err;
+
+  dbus_error_init(&err);
+  server = dbus_server_listen(spi_global_app_data->app_bus_addr, &err);
+  if (server == NULL)
+  {
+    g_warning (_("atk-bridge: Couldn't listen on dbus server: %s"), err.message);
+    dbus_error_init (&err);
+    spi_global_app_data->app_bus_addr [0] = '\0';
+    g_main_context_unref (spi_global_app_data->main_context);
+    spi_global_app_data->main_context = NULL;
+    return -1;
+  }
+
+  dbus_server_setup_with_g_main(server, NULL);
+  dbus_server_set_new_connection_function(server, new_connection_cb, NULL, NULL);
+
+  spi_global_app_data->server = server;
+#endif
+
+  return 0;
+}
+
+
 gchar *atspi_dbus_name = NULL;
 static gboolean atspi_no_register = FALSE;
 
@@ -575,6 +617,8 @@ remove_events (const char *bus_name, const char *event)
           list = list->next;
         }
     }
+
+  g_strfreev (remove_data);
 }
 
 static void
@@ -622,6 +666,7 @@ signal_filter (DBusConnection *bus, DBusMessage *message, void *user_data)
  *
  * - DRoute for routing message to their accessible objects.
  * - Event handlers for emmitting signals on specific ATK events.
+ * - setup the bus for p2p communication
  * - Application registration with the AT-SPI registry.
  *
  */
@@ -656,6 +701,7 @@ adaptor_init (gint * argc, gchar ** argv[])
   g_option_context_set_ignore_unknown_options (opt, TRUE);
   if (!g_option_context_parse (opt, argc, argv, &err))
     g_warning ("AT-SPI Option parsing failed: %s\n", err->message);
+  g_option_context_free (opt);
 
   /* Allocate global data and do ATK initializations */
   spi_global_app_data = g_new0 (SpiBridge, 1);
@@ -688,8 +734,13 @@ adaptor_init (gint * argc, gchar ** argv[])
         }
     }
 
-  dbus_connection_setup_with_g_main (spi_global_app_data->bus,
-                                     g_main_context_default ());
+#ifndef DISABLE_P2P
+  spi_global_app_data->main_context = g_main_context_new ();
+#else
+  spi_global_app_data->main_context = NULL;
+#endif
+
+  dbus_connection_setup_with_g_main (spi_global_app_data->bus, NULL);
 
   /* Hook our plug-and socket functions */
   install_plug_hooks ();
@@ -705,7 +756,7 @@ adaptor_init (gint * argc, gchar ** argv[])
 
   /* Register droute for routing AT-SPI messages */
   spi_global_app_data->droute =
-    droute_new (spi_global_app_data->bus);
+    droute_new ();
 
   treepath = droute_add_one (spi_global_app_data->droute,
                              "/org/a11y/atspi/cache", spi_global_cache);
@@ -743,6 +794,9 @@ adaptor_init (gint * argc, gchar ** argv[])
   spi_initialize_text (accpath);
   spi_initialize_value (accpath);
 
+  droute_context_register (spi_global_app_data->droute,
+                           spi_global_app_data->bus);
+
   /* Register methods to send D-Bus signals on certain ATK events */
   spi_atk_register_event_listeners ();
 
@@ -757,6 +811,8 @@ adaptor_init (gint * argc, gchar ** argv[])
   else
     get_registered_event_listeners (spi_global_app_data);
 
+  setup_bus();
+
   return 0;
 }
 
